import logging
from collections.abc import Iterable
from pathlib import Path
from time import perf_counter
from typing import Dict

import h5py
import numpy as np
import pandas as pd
import yaml
from utils_find_1st import cmp_equal, find_1st

from agora.io.bridge import BridgeH5

#################### Dynamic version ##################################


def load_attributes(file: str, group="/"):
    """
    Load the metadata from an h5 file and convert to a dictionary, including the "parameters" field which is stored as YAML.

    Parameters
    ----------
    file: str
        Name of the h5 file
    group: str, optional
        The group in the h5 file from which to read the data
    """
    # load the metadata, stored as attributes, from the h5 file and return as a dictionary
    with h5py.File(file, "r") as f:
        meta = dict(f[group].attrs.items())
    if "parameters" in meta:
        # convert from yaml format into dict
        meta["parameters"] = yaml.safe_load(meta["parameters"])
    return meta


class DynamicWriter:
    """Provides a parent class for all writers."""

    # a dict giving for each dataset a tuple, comprising the dataset's maximum size, as a 2D tuple, and its type
    data_types = {}
    # the group in the h5 file to write to
    group = ""
    # compression info
    compression = "gzip"
    compression_opts = 9
    metadata = None

    def __init__(self, file: str):
        self.file = file
        # the metadata is stored as attributes in the h5 file
        if Path(file).exists():
            self.metadata = load_attributes(file)

    def _log(self, message: str, level: str = "warn"):
        # Log messages in the corresponding level
        logger = logging.getLogger("aliby")
        getattr(logger, level)(f"{self.__class__.__name__}: {message}")

    def _append(self, data, key, hgroup):
        """
        Append data to existing dataset in the h5 file otherwise create a new one.

        Parameters
        ----------
        data
            Data to be written, typically a numpy array
        key: str
            Name of dataset
        hgroup: str
            Destination group in the h5 file
        """
        try:
            n = len(data)
        except Exception as e:
            logging.debug(
                "DynamicWriter: Attributes have no length: {}".format(e)
            )
            n = 1
        if key in hgroup:
            # append to existing dataset
            try:
                # FIXME This is broken by bugged mother-bud assignment
                dset = hgroup[key]
                dset.resize(dset.shape[0] + n, axis=0)
                dset[-n:] = data
            except Exception as e:
                logging.debug(
                    "DynamicWriter: Inconsistency between dataset shape and new empty data: {}".format(
                        e
                    )
                )
        else:
            # create new dataset
            # TODO Include sparsity check
            max_shape, dtype = self.datatypes[key]
            shape = (n,) + max_shape[1:]
            hgroup.create_dataset(
                key,
                shape=shape,
                maxshape=max_shape,
                dtype=dtype,
                compression=self.compression,
                compression_opts=self.compression_opts
                if self.compression is not None
                else None,
            )
            # write all data, signified by the empty tuple
            hgroup[key][()] = data

    def _overwrite(self, data, key, hgroup):
        """
        Delete and then replace existing dataset in h5 file.

        Parameters
        ----------
        data
            Data to be written, typically a numpy array
        key: str
            Name of dataset
        hgroup: str
            Destination group in the h5 file
        """
        # We do not append to mother_assign; raise error if already saved
        data_shape = np.shape(data)
        max_shape, dtype = self.datatypes[key]
        # delete existing data
        if key in hgroup:
            del hgroup[key]
        # write new data
        hgroup.require_dataset(
            key,
            shape=data_shape,
            dtype=dtype,
            compression=self.compression,
        )
        # write all data, signified by the empty tuple
        hgroup[key][()] = data

    def write(self, data: dict, overwrite: list, meta: dict = {}):
        """
        Write data and metadata to h5 file.

         Parameters
        ----------
        data: dict
            A dict of datasets and data
        overwrite: list of str
            A list of datasets to overwrite
        meta: dict, optional
            Metadata to be written as attributes of the h5 file
        """
        with h5py.File(self.file, "a") as store:
            # open group, creating if necessary
            hgroup = store.require_group(self.group)
            # write data
            for key, value in data.items():
                # only save data with a pre-defined data-type
                if key not in self.datatypes:
                    raise KeyError(f"No defined data type for key {key}")
                else:
                    try:
                        if key.startswith("attrs/"):
                            # metadata
                            key = key.split("/")[1]
                            hgroup.attrs[key] = value
                        elif key in overwrite:
                            # delete and replace existing dataset
                            self._overwrite(value, key, hgroup)
                        else:
                            # append or create new dataset
                            self._append(value, key, hgroup)
                    except Exception as e:
                        print(key, value)
                        self._log(
                            f"{key}:{value} could not be written: {e}", "error"
                        )
            # write metadata
            for key, value in meta.items():
                hgroup.attrs[key] = value


##################### Special instances #####################
class TilerWriter(DynamicWriter):
    """Write data stored in a Tiler instance to h5 files."""

    datatypes = {
        "trap_locations": ((None, 2), np.uint16),
        "drifts": ((None, 2), np.float32),
        "attrs/tile_size": ((1,), np.uint16),
        "attrs/max_size": ((1,), np.uint16),
    }
    group = "trap_info"

    def write(self, data: dict, overwrite: list, tp: int, meta: dict = {}):
        """
        Write data for time points that have none.

        Parameters
        ----------
        data: dict
            A dict of datasets and data
        overwrite: list of str
            A list of datasets to overwrite
        tp: int
            The time point of interest
        meta: dict, optional
            Metadata to be written as attributes of the h5 file
        """
        skip = False
        # append to h5 file
        with h5py.File(self.file, "a") as store:
            # open group, creating if necessary
            hgroup = store.require_group(self.group)
            # find xy drift for each time point as proof that it has already been processed
            nprev = hgroup.get("drifts", None)
            if nprev and tp < nprev.shape[0]:
                # data already exists
                print(f"Tiler: Skipping timepoint {tp}")
                skip = True
        if not skip:
            super().write(data=data, overwrite=overwrite, meta=meta)


class LinearBabyWriter(DynamicWriter):
    """
    Write data stored in a Baby instance to h5 files.

    Assumes the edgemasks are of form ((None, tile_size, tile_size), bool).
    """

    compression = "gzip"
    _default_tile_size = 117
    datatypes = {
        "centres": ((None, 2), np.uint16),
        "position": ((None,), np.uint16),
        "angles": ((None,), h5py.vlen_dtype(np.float32)),
        "radii": ((None,), h5py.vlen_dtype(np.float32)),
        "edgemasks": ((None, _default_tile_size, _default_tile_size), bool),
        "ellipse_dims": ((None, 2), np.float32),
        "cell_label": ((None,), np.uint16),
        "trap": ((None,), np.uint16),
        "timepoint": ((None,), np.uint16),
        "mother_assign_dynamic": ((None,), np.uint16),
        "volumes": ((None,), np.float32),
    }
    group = "cell_info"

    def write(
        self, data: dict, overwrite: list, tp: int = None, meta: dict = {}
    ):
        """
        Check data does not exist before writing.

        Parameters
        ----------
        data: dict
            A dict of datasets and data
        overwrite: list of str
            A list of datasets to overwrite
        tp: int
            The time point of interest
        meta: dict, optional
            Metadata to be written as attributes of the h5 file
        """
        with h5py.File(self.file, "a") as store:
            hgroup = store.require_group(self.group)
            available_tps = hgroup.get("timepoint", None)
            # write data
            if not available_tps or tp not in np.unique(available_tps[()]):
                super().write(data, overwrite)
            else:
                # data already exists
                print(f"BabyWriter: Skipping tp {tp}")
            # write metadata
            for key, value in meta.items():
                hgroup.attrs[key] = value


class StateWriter(DynamicWriter):
    """
    Write information summarising the current state of the pipeline - the 'last_state' dataset in the h5 file.

    MOVEDatatypes are specified with the first variable specifying the number of traps and the other specifying the shape of the object.

    """

    datatypes = {
        # the highest cell label assigned for each time point
        "max_lbl": ((None, 1), np.uint16),
        # how far back we go for tracking
        "tp_back": ((None, 1), np.uint16),
        # trap labels
        "trap": ((None, 1), np.int16),
        # cell labels
        "cell_lbls": ((None, 1), np.uint16),
        # previous cell features for tracking
        "prev_feats": ((None, None), np.float32),
        # number of images for which a cell has been present
        "lifetime": ((None, 2), np.uint16),
        # probability of a mother-bud relationship given a bud
        "p_was_bud": ((None, 2), np.float32),
        # probability of a mother-bud relationship given a mother
        "p_is_mother": ((None, 2), np.float32),
        # cumulative matrix, over time, of bud assignments
        "ba_cum": ((None, None), np.float32),
    }
    group = "last_state"
    compression = "gzip"

    @staticmethod
    def format_field(states: list, field: str):
        """Flatten a field in the states list to save as an h5 dataset."""
        fields = [pos_state[field] for pos_state in states]
        return fields

    @staticmethod
    def format_values_tpback(states: list, val_name: str):
        """Unpacks a dict of state data into tp_back, trap, value."""

        # store results as a list of tuples
        lbl_tuples = [
            (tp_back, trap, cell_label)
            for trap, state in enumerate(states)
            for tp_back, value in enumerate(state[val_name])
            for cell_label in value
        ]
        # unpack list of tuples to define variables
        if len(lbl_tuples):
            tp_back, trap, value = zip(*lbl_tuples)
        else:
            # set as empty lists
            tp_back, trap, value = [
                [[] for _ in states[0][val_name]] for _ in range(3)
            ]
        return tp_back, trap, value

    @staticmethod
    def format_values_traps(states: list, val_name: str):
        """Format either lifetime, p_was_bud, or p_is_mother variables as a list."""
        formatted = np.array(
            [
                (trap, clabel_val)
                for trap, state in enumerate(states)
                for clabel_val in state[val_name]
            ]
        )
        return formatted

    @staticmethod
    def pad_if_needed(array: np.ndarray, pad_size: int):
        """Pad a 2D array with zeros for large indices."""
        padded = np.zeros((pad_size, pad_size)).astype(float)
        length = len(array)
        padded[:length, :length] = array
        return padded

    def format_states(self, states: list):
        """Re-format state data into a dict of lists, with one element per per list per state."""
        formatted_state = {"max_lbl": [state["max_lbl"] for state in states]}
        tp_back, trap, cell_label = self.format_values_tpback(
            states, "cell_lbls"
        )
        _, _, prev_feats = self.format_values_tpback(states, "prev_feats")
        # store lists in a dict
        formatted_state["tp_back"] = tp_back
        formatted_state["trap"] = trap
        formatted_state["cell_lbls"] = cell_label
        formatted_state["prev_feats"] = np.array(prev_feats)
        # one entry per cell label - tp_back independent
        for val_name in ("lifetime", "p_was_bud", "p_is_mother"):
            formatted_state[val_name] = self.format_values_traps(
                states, val_name
            )
        bacum_max = max([len(state["ba_cum"]) for state in states])
        formatted_state["ba_cum"] = np.array(
            [
                self.pad_if_needed(state["ba_cum"], bacum_max)
                for state in states
            ]
        )
        return formatted_state

    def write(self, data: dict, overwrite: list, tp: int = 0):
        """Write the current state of the pipeline."""
        if len(data):
            last_tp = 0
            try:
                with h5py.File(self.file, "r") as f:
                    gr = f.get(self.group, None)
                    if gr:
                        last_tp = gr.attrs.get("tp", 0)
                if tp == 0 or tp > last_tp:
                    # write
                    formatted_data = self.format_states(data)
                    super().write(data=formatted_data, overwrite=overwrite)
                    with h5py.File(self.file, "a") as f:
                        # record that data for the timepoint has been written
                        f[self.group].attrs["tp"] = tp
                elif tp > 0 and tp <= last_tp:
                    # data already present
                    print(f"StateWriter: Skipping timepoint {tp}")
            except Exception as e:
                raise (e)
        else:
            print("Skipping overwriting: no data")


#################### Extraction version ###############################
class Writer(BridgeH5):
    """
    Class to transform data into compatible structures.
    Used by Extractor and Postprocessor within the pipeline."""

    def __init__(self, filename, flag=None, compression="gzip"):
        """
        Initialise write.

        Parameters
        ----------
        filename: str
            Name of file to write into
        flag: str, default=None
            Flag to pass to the default file reader. If None the file remains closed.
        compression: str, default="gzip"
            Compression method passed on to h5py writing functions (only used for dataframes and other array-like data).
        """
        super().__init__(filename, flag=flag)
        self.compression = compression

    def write(
        self,
        path: str,
        data: Iterable = None,
        meta: dict = {},
        overwrite: str = None,
    ):
        """
        Write data and metadata to a particular path in the h5 file.

        Parameters
        ----------
        path : str
            Path inside h5 file into which to write.
        data : Iterable, optional
        meta : dict, optional
        overwrite: str, optional
        """
        self.id_cache = {}
        with h5py.File(self.filename, "a") as f:
            # Alan, haven't we already opened the h5 file through BridgeH5's init?
            if overwrite == "overwrite":  # TODO refactor overwriting
                if path in f:
                    del f[path]
            logging.debug(
                "{} {} to {} and {} metadata fields".format(
                    overwrite, type(data), path, len(meta)
                )
            )
            # write data
            if data is not None:
                self.write_dset(f, path, data)
            # write metadata
            if meta:
                for attr, metadata in meta.items():
                    self.write_meta(f, path, attr, data=metadata)

    def write_dset(self, f: h5py.File, path: str, data: Iterable):
        """Write data in different ways depending on its type to an open h5 file."""
        # data is a datafram
        if isinstance(data, pd.DataFrame):
            self.write_pd(f, path, data, compression=self.compression)
        # data is a multi-index dataframe
        elif isinstance(data, pd.MultiIndex):
            # TODO: benchmark I/O speed when using compression
            self.write_index(f, path, data)  # , compression=self.compression)
        # data is a dictionary of dataframes
        elif isinstance(data, Dict) and np.all(
            [isinstance(x, pd.DataFrame) for x in data.values]
        ):
            for k, df in data.items():
                self.write_dset(f, path + f"/{k}", df)
        # data is an iterable
        elif isinstance(data, Iterable):
            self.write_arraylike(f, path, data)
        # data is a float or integer
        else:
            self.write_atomic(data, f, path)

    def write_meta(self, f: h5py.File, path: str, attr: str, data: Iterable):
        """Write metadata to an open h5 file."""
        obj = f.require_group(path)
        obj.attrs[attr] = data

    @staticmethod
    def write_arraylike(f: h5py.File, path: str, data: Iterable, **kwargs):
        """Write an iterable."""
        if path in f:
            del f[path]
        narray = np.array(data)
        if narray.any():
            chunks = (1, *narray.shape[1:])
        else:
            chunks = None
        # create dset
        dset = f.create_dataset(
            path,
            shape=narray.shape,
            chunks=chunks,
            dtype="int",
            compression=kwargs.get("compression", None),
        )
        # add data to dset
        dset[()] = narray

    @staticmethod
    def write_index(f, path, pd_index, **kwargs):
        """Write a multi-index dataframe."""
        f.require_group(path)  # TODO check if we can remove this
        for i, name in enumerate(pd_index.names):
            ids = pd_index.get_level_values(i)
            id_path = path + "/" + name
            f.create_dataset(
                name=id_path,
                shape=(len(ids),),
                dtype="uint16",
                compression=kwargs.get("compression", None),
            )
            indices = f[id_path]
            indices[()] = ids

    def write_pd(self, f, path, df, **kwargs):
        """Write a dataframe."""
        values_path = (
            path + "values" if path.endswith("/") else path + "/values"
        )
        if path not in f:

            # create dataset and write data
            max_ncells = 2e5
            max_tps = 1e3
            f.create_dataset(
                name=values_path,
                shape=df.shape,
                # chunks=(min(df.shape[0], 1), df.shape[1]),
                # dtype=df.dtypes.iloc[0], This is making NaN in ints into negative vals
                dtype="float",
                maxshape=(max_ncells, max_tps),
                compression=kwargs.get("compression", None),
            )
            dset = f[values_path]
            dset[()] = df.values

            # create dateset and write indices
            if not len(df):  # Only write more if not empty
                return None

            for name in df.index.names:
                indices_path = "/".join((path, name))
                f.create_dataset(
                    name=indices_path,
                    shape=(len(df),),
                    dtype="uint16",  # Assuming we'll always use int indices
                    chunks=True,
                    maxshape=(max_ncells,),
                )
                dset = f[indices_path]
                dset[()] = df.index.get_level_values(level=name).tolist()
            # create dataset and write columns
            if (
                df.columns.dtype == int
                or df.columns.dtype == np.dtype("uint")
                or df.columns.name == "timepoint"
            ):
                tp_path = path + "/timepoint"
                f.create_dataset(
                    name=tp_path,
                    shape=(df.shape[1],),
                    maxshape=(max_tps,),
                    dtype="uint16",
                )
                tps = list(range(df.shape[1]))
                f[tp_path][tps] = tps
            else:
                f[path].attrs["columns"] = df.columns.tolist()
        else:

            # path exists
            dset = f[values_path]

            # filter out repeated timepoints
            new_tps = set(df.columns)
            if path + "/timepoint" in f:
                new_tps = new_tps.difference(f[path + "/timepoint"][()])
            df = df[new_tps]

            if (
                not hasattr(self, "id_cache")
                or df.index.nlevels not in self.id_cache
            ):
                # use cache dict to store previously obtained indices
                self.id_cache[df.index.nlevels] = {}
                existing_ids = self.get_existing_ids(
                    f, [path + "/" + x for x in df.index.names]
                )
                # split indices in existing and additional
                new = df.index.tolist()
                if df.index.nlevels == 1:
                    # cover cases with a single index
                    new = [(x,) for x in df.index.tolist()]
                (
                    found_multis,
                    self.id_cache[df.index.nlevels]["additional_multis"],
                ) = self.find_ids(
                    existing=existing_ids,
                    new=new,
                )
                found_indices = np.array(
                    locate_indices(existing_ids, found_multis)
                )

                # sort indices for h5 indexing
                incremental_existing = np.argsort(found_indices)
                self.id_cache[df.index.nlevels][
                    "found_indices"
                ] = found_indices[incremental_existing]
                self.id_cache[df.index.nlevels]["found_multi"] = found_multis[
                    incremental_existing
                ]

            existing_values = df.loc[
                [
                    _tuple_or_int(x)
                    for x in self.id_cache[df.index.nlevels]["found_multi"]
                ]
            ].values
            new_values = df.loc[
                [
                    _tuple_or_int(x)
                    for x in self.id_cache[df.index.nlevels][
                        "additional_multis"
                    ]
                ]
            ].values
            ncells, ntps = f[values_path].shape

            # add found cells
            dset.resize(dset.shape[1] + df.shape[1], axis=1)
            dset[:, ntps:] = np.nan

            # TODO refactor this indices sorting. Could be simpler
            found_indices_sorted = self.id_cache[df.index.nlevels][
                "found_indices"
            ]

            # case when all labels are new
            if found_indices_sorted.any():
                # h5py does not allow bidimensional indexing,
                # so we have to iterate over the columns
                for i, tp in enumerate(df.columns):
                    dset[found_indices_sorted, tp] = existing_values[:, i]
            # add new cells
            n_newcells = len(
                self.id_cache[df.index.nlevels]["additional_multis"]
            )
            dset.resize(dset.shape[0] + n_newcells, axis=0)
            dset[ncells:, :] = np.nan

            for i, tp in enumerate(df.columns):
                dset[ncells:, tp] = new_values[:, i]

            # save indices
            for i, name in enumerate(df.index.names):
                tmp = path + "/" + name
                dset = f[tmp]
                n = dset.shape[0]
                dset.resize(n + n_newcells, axis=0)
                dset[n:] = (
                    self.id_cache[df.index.nlevels]["additional_multis"][:, i]
                    if len(
                        self.id_cache[df.index.nlevels][
                            "additional_multis"
                        ].shape
                    )
                    > 1
                    else self.id_cache[df.index.nlevels]["additional_multis"]
                )

            tmp = path + "/timepoint"
            dset = f[tmp]
            n = dset.shape[0]
            dset.resize(n + df.shape[1], axis=0)
            dset[n:] = df.columns.tolist()

    @staticmethod
    def get_existing_ids(f, paths):
        """Fetch indices and convert them to a (nentries, nlevels) ndarray."""
        return np.array([f[path][()] for path in paths]).T

    @staticmethod
    def find_ids(existing, new):
        """Compare two tuple sets and return the intersection and difference (elements in the 'new' set not in 'existing')."""
        set_existing = set([tuple(*x) for x in zip(existing.tolist())])
        existing_cells = np.array(list(set_existing.intersection(new)))
        new_cells = np.array(list(set(new).difference(set_existing)))
        return existing_cells, new_cells


def locate_indices(existing, new):
    if new.any():
        if new.shape[1] > 1:
            return [
                find_1st(
                    (existing[:, 0] == n[0]) & (existing[:, 1] == n[1]),
                    True,
                    cmp_equal,
                )
                for n in new
            ]
        else:
            return [
                find_1st(existing[:, 0] == n, True, cmp_equal) for n in new
            ]
    else:
        return []


def _tuple_or_int(x):
    """Convert tuple to int if it only contains one value."""
    if len(x) == 1:
        return x[0]
    else:
        return x



import csv
import io
import operator
import re
import typing as t
from collections import Counter
from datetime import datetime
from pathlib import Path, PosixPath

import numpy as np
from logfile_parser import Parser
from omero.gateway import BlitzGateway, TagAnnotationWrapper, _DatasetWrapper
from tqdm import tqdm


class OmeroExplorer:
    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        min_date: t.Tuple[int, int, int] = (2020, 6, 1),
    ):
        self.conn = BlitzGateway(user, password, host=host)
        self.conn.connect()

        self.min_date = min_date
        self.backups = {}
        self.removed = []

    @property
    def cache(self):
        if not hasattr(self, "_cache"):
            self._cache = {v.id: get_annotsets(v) for v in self.dsets}
        return self._cache

    @property
    def raw_log(self):
        return {k: v["log"] for k, v in self.cache.items()}

    @property
    def raw_log_end(self):
        if not hasattr(self, "_raw_log_end"):
            self._raw_log_end = {d.id: get_logfile(d) for d in self.dsets}
        return self._raw_log_end

    @property
    def log(self):
        return {k: parse_annot(v, "log") for k, v in self.raw_log.items()}

    @property
    def raw_acq(self):
        return {k: v["acq"] for k, v in self.cache.items()}

    @property
    def acq(self):
        return {k: parse_annot(v, "acq") for k, v in self.raw_acq.items()}

    def load(self, min_id=0, min_date=None):
        """
        :min_id: int
        :min_date: tuple
        """
        if min_date is None:
            min_date = self.min_date
        self._dsets_bak = [
            d for d in self.conn.getObjects("Dataset") if d.getId() > min_id
        ]

        if min_date:
            if len(min_date) < 3:
                min_date = min_date + tuple(
                    [1 for _ in range(3 - len(min_date))]
                )
            min_date = datetime(*min_date)

            # sort by dates
            dates = [d.getDate() for d in self._dsets_bak]
            self._dsets_bak[:] = [
                a
                for _, a in sorted(
                    zip(dates, self._dsets_bak), key=lambda x: x[0]
                )
            ]

            self._dsets_bak = [
                d for d in self._dsets_bak if d.getDate() >= min_date
            ]

        self.dsets = self._dsets_bak
        self.n_dsets

    def image_ids(self):
        return {
            dset.getId(): [im.getId() for im in dset.listChildren()]
            for dset in self.dsets
        }

    def dset(self, n):
        try:
            return [x for x in self.dsets if x.id == n][0]
        except Exception as e:
            print(f"Could not fetch all data xsets: {e}")

    def channels(self, setkey, present=True):
        """
        :setkey: str indicating a set of channels
        :present: bool indicating whether the search should or shold not be in the dset
        """
        self.dsets = [
            v for v in self.acqs.values() if present == has_channels(v, setkey)
        ]
        self.n_dsets

    def update_cache(self):
        if not hasattr(self, "acq") or not hasattr(self, "log"):
            for attr in ["acq", "log"]:
                print("Updating raw ", attr)
                setattr(
                    self,
                    "raw_" + attr,
                    {v.id: get_annotsets(v)[attr] for v in self.dsets},
                )
                setattr(
                    self,
                    attr,
                    {
                        v.id: parse_annot(
                            getattr(self, "raw_" + attr)[v.id], attr
                        )
                        for v in self.dsets
                    },
                )
        else:

            for attr in ["acq", "log", "raw_acq", "raw_log"]:
                setattr(
                    self,
                    attr,
                    {i.id: getattr(self, attr)[i.id] for i in self.dsets},
                )

    @property
    def dsets(self):
        if not hasattr(self, "_dsets"):
            self._dsets = self.load()

        return self._dsets

    @dsets.setter
    def dsets(self, dsets):
        if hasattr(self, "_dsets"):
            if self._dsets is None:
                self._dsets = []
            self.removed += [
                x for x in self._dsets if x.id not in [y.id for y in dsets]
            ]

        self._dsets = dsets

    def tags(self, tags, present=True):
        """
        :setkey str tags to filter
        """
        if type(tags) is not list:
            tags = [str(tags)]

        self.dsets = [
            v for v in self.dsets if present == self.has_tags(v, tags)
        ]
        self.n_dsets

    @property
    def all_tags(self):
        if not hasattr(self, "_tags"):
            self._tags = {
                d.id: [
                    x.getValue()
                    for x in d.listAnnotations()
                    if isinstance(x, TagAnnotationWrapper)
                ]
                for d in self.dsets
            }
        return self._tags

    def get_timepoints(self):
        self.image_wrappers = {
            d.id: list(d.listChildren())[0] for d in self.dsets
        }

        return {k: i.getSizeT() for k, i in self.image_wrappers.items()}

    def timepoints(self, n, op="greater"):
        "Filter experiments using the number of timepoints"
        op = operator.gt if op == "greater" else operator.le
        self._timepoints = self.get_timepoints()

        self.dsets = [
            v for v in tqdm(self.dsets) if op(self._timepoints[v.id], n)
        ]

    def microscope(self, microscope):
        self.microscopes = {
            dset.id: self.get_microscope(self.log[dset.id])
            for dset in self.dsets
        }

        self.n_dsets

    def get_microscope(self, parsed_log):
        return parsed_log["microscope"]

    def reset(self, backup_id=None):
        self.dsets = self.backups.get(backup_id, self._dsets_bak)
        self.n_dsets

    def backup(self, name):
        self.backups[name] = self.dsets

    def reset_backup(self, name):
        self.dsets = self.backups[name]

    @staticmethod
    def is_complete(logfile: str):
        return logfile.endswith("Experiment completed\r\r\n")

    @staticmethod
    def count_errors(logfile: str):
        return re.findall("ERROR CAUGHT", logfile)

    @staticmethod
    def count_drift_alert(logfile: str):
        return re.findall("High drift alert!", logfile)

    @staticmethod
    def is_interrupted(logfile: str):
        return "Experiment stopped by user" in logfile

    @property
    def n_dsets(self):
        print("{} datasets.".format(len(self.dsets)))

    @property
    def desc(self):
        for d in self.dsets:
            print(
                "{}\t{}\t{}\t{}".format(
                    d.getDate().strftime("%x"),
                    d.getId(),
                    d.getName(),
                    d.getDescription(),
                )
            )

    @property
    def ids(self):
        return [d.getId() for d in self.dsets]

    def get_ph_params(self):
        t = [
            {
                ch: [exp, v]
                for ch, exp, v in zip(
                    j["channel"], j["exposure"], j["voltage"]
                )
                if ch in {"GFPFast", "pHluorin405"}
            }
            for j in [i["channels"] for i in self.acqs]
        ]

        ph_param_pairs = [
            (tuple(x.values())) for x in t if np.all(list(x.values()))
        ]

        return Counter([str(x) for x in ph_param_pairs])

    def find_duplicate_candidates(self, days_tol=2):
        # Find experiments with the same name or Aim and from similar upload dates
        # and group them for cleaning
        pass

    def return_completed(self, kind="complete"):
        return {
            k: getattr(self, f"is_{kind}")(v.get("log", ""))
            for k, v in self.cache.items()
        }

    def dset_count(
        self,
        dset: t.Union[int, _DatasetWrapper],
        kind: str = "errors",
        norm: bool = True,
    ):
        if isinstance(dset, int):
            dset = self.conn.getObject("Dataset", dset)

        total_images_tps = sum([im.getSizeT() for im in dset.listChildren()])

        return len(
            getattr(self, f"count_{kind}")(
                self.cache[dset.getId()].get("log", ""), norm=norm
            )
        ) / (norm * total_images_tps)

    def count_in_log(self, kind="errors", norm: bool = True):
        return {
            k: self.dset_count(k, kind=kind, norm=norm)
            for k, v in self.cache.items()
        }

    @property
    def complete(self):
        self.completed = {
            k: self.is_complete(v.get("log", ""))
            for k, v in self.cache.items()
        }
        self.dsets = [dset for dset in self.dsets if self.completed[dset.id]]
        return self.n_dsets

    def save(self, fname):
        with open(fname + ".tsv", "w") as f:
            writer = csv.writer(f, delimiter="\t")
            for d in self.dsets:
                writer.writerow(
                    [
                        d.getDate().strftime("%x"),
                        d.getId(),
                        d.getName(),
                        d.getDescription(),
                    ]
                )

    @property
    def positions(self):
        return {x.id: len(list(x.listChildren())) for x in self.dsets}

    def has_tags(self, d, tags):
        if set(tags).intersection(self.all_tags[d.id]):
            return True


def convert_to_hours(delta):
    total_seconds = delta.total_seconds()
    hours = int(total_seconds // 3600)
    return hours


class Argo(OmeroExplorer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def annot_from_dset(dset, kind):
    v = [x for x in dset.listAnnotations() if hasattr(x, "getFileName")]
    infname = kind if kind == "log" else kind.title()
    try:
        acqfile = [x for x in v if x.getFileName().endswith(infname + ".txt")][
            0
        ]
        decoded = list(acqfile.getFileInChunks())[0].decode("utf-8")
        acq = parse_annot(decoded, kind)
    except Exception as e:
        print(f"Conversion from acquisition file failed: {e}")
        return {}
    return acq


def check_channels(acq, channels, _all=True):
    shared_channels = set(acq["channels"]["channel"]).intersection(channels)

    condition = False
    if _all:
        if len(shared_channels) == len(channels):
            condition = True
    else:
        if len(shared_channels):
            condition = True

    return condition


def get_chs(exptype):
    # TODO Documentation
    exptypes = {
        "dual_ph": ("GFP", "pHluorin405", "mCherry"),
        "ph": ("GFP", "pHluorin405"),
        "dual": ("GFP", "mCherry"),
        "mCherry": ("mCherry"),
    }
    return exptypes[exptype]


def load_annot_from_cache(exp_id, cache_dir="cache/"):
    # TODO Documentation
    if type(cache_dir) is not PosixPath:
        cache_dir = Path(cache_dir)

    annot_sets = {}
    for fname in cache_dir.joinpath(exp_id).rglob("*.txt"):
        fmt = fname.name[:3]
        with open(fname, "r") as f:
            annot_sets[fmt] = f.read()

    return annot_sets


def get_annot(annot_sets, fmt):
    """
    Get parsed annotation file
    """
    str_io = annot_sets.get(fmt, None)
    return parse_annot(str_io, fmt)


def parse_annot(str_io, fmt):
    parser = Parser("multiDGUI_" + fmt + "_format")
    return parser.parse(io.StringIO(str_io))


def get_annotsets(dset):
    annot_files = [
        annot.getFile()
        for annot in dset.listAnnotations()
        if hasattr(annot, "getFile")
    ]
    annot_sets = {
        ftype[:-4].lower(): annot
        for ftype in ("log.txt", "Acq.txt", "Pos.txt")
        for annot in annot_files
        if ftype in annot.getName()
    }
    annot_sets = {
        key: list(a.getFileInChunks())[0].decode("utf-8")
        for key, a in annot_sets.items()
    }
    return annot_sets


def load_acq(dset):
    # TODO Documentation
    try:
        acq = annot_from_dset(dset, kind="acq")
        return acq
    except Exception as e:
        print(f"dset{dset.getId()}failed acq loading: {e}")
        return False


def has_channels(dset, exptype):
    # TODO Documentation
    acq = load_acq(dset)
    if acq:
        return check_channels(acq, get_chs(exptype))
    else:
        return


def get_logfile(dset):
    # TODO Documentation
    annot_file = [
        annot.getFile()
        for annot in dset.listAnnotations()
        if hasattr(annot, "getFile")
        and annot.getFileName().endswith("log.txt")
    ][0]
    return list(annot_file.getFileInChunks())[-1].decode("utf-8")



"""
ImageViewer class, used to look at individual or multiple traps over time.


Example of usage:

fpath = "/home/alan/Documents/dev/skeletons/scripts/data/16543_2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01/URA8_young018.h5"

trap_id = 9
trange = list(range(0, 30))
ncols = 8

riv = remoteImageViewer(fpath)
riv.plot_labelled_trap(trap_id, trange, [0], ncols=ncols)

"""

import re
import typing as t

import h5py
from abc import ABC
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from PIL import Image
from skimage.morphology import dilation

from agora.io.cells import Cells
from agora.io.writer import load_attributes
from aliby.io.image import dispatch_image
from aliby.tile.tiler import Tiler, TilerParameters
from aliby.utils.plot import stretch_image

default_colours = {
    "Brightfield": "Greys_r",
    "GFP": "Greens_r",
    "mCherry": "Reds_r",
    "cell_label": sns.color_palette("Paired", as_cmap=True),
}


def custom_imshow(a, norm=None, cmap=None, *args, **kwargs):
    """
    Wrapper on plt.imshow function.
    """
    if cmap is None:
        cmap = "Greys_r"
    return plt.imshow(
        a,
        *args,
        cmap=cmap,
        interpolation=None,
        interpolation_stage="rgba",
        **kwargs,
    )


class localImageViewer:
    """
    Fast access to Images segmented locally without tiling
    from image.h5 objects.
    """

    def __init__(self, h5file, data_source=None):
        self._hdf = h5py.File(h5file)
        self.positions = list(self._hdf.keys())
        self.current_position = self.positions[0]

    def plot_position(self, channel=0, tp=0, z=0, stretch=True):
        pixvals = self._hdf[self.current_position][channel, tp, ..., z]
        if stretch:
            minval = np.percentile(pixvals, 0.5)
            maxval = np.percentile(pixvals, 99.5)
            pixvals = np.clip(pixvals, minval, maxval)
            pixvals = ((pixvals - minval) / (maxval - minval)) * 255

        Image.fromarray(pixvals.astype(np.uint8))


class BaseImageViewer(ABC):
    def __init__(self, fpath):

        self._fpath = fpath
        attrs = load_attributes(fpath)
        self._logfiles_meta = {}
        self._logfiles_meta["channels"] = attrs["channels/channel"]

        self.image_id = attrs.get("image_id")

        assert self.image_id is not None, "No valid image_id found in metadata"

    @property
    def shape(self):
        return self.tiler.image.shape

    @property
    def ntraps(self):
        return self.cells.ntraps

    @property
    def max_labels(self):
        # Print max cell label in whole experiment
        return [max(x) for x in self.cells.labels]

    def labels_at_time(self, tp: int):
        # Print  cell label at a given time-point
        return self.cells.labels_at_time(tp)


class LocalImageViewer(BaseImageViewer):
    """
    Tool to generate figures from local files, either zarr or files organised
    in directories.
    """

    def __init__(self, results_path: str, data_path: str):
        super().__init__(results_path)

        from aliby.io.image import ImageDir, ImageZarr

        self._image_class = (
            ImageZarr if data_path.endswith(".zar") else ImageDir
        )

        with dispatch_image(data_path)(data_path) as image:
            self.tiler = Tiler(
                image.data,
                self._meta if hasattr(self, "_meta") else self._logfiles_meta,
                TilerParameters.default(),
            )

        self.cells = Cells.from_source(results_path)


class remoteImageViewer(BaseImageViewer):
    """
    This ImageViewer combines fetching remote images with tiling and outline display.
    """

    def __init__(
        self,
        results_path: str,
        server_info: t.Dict[str, str],
    ):
        self.super().__init__(results_path)

        from aliby.io.omero import Image as OImage

        self._image_class = OImage

        self._server_info = (
            server_info or attrs["parameters"]["general"]["server_info"]
        )

        with dispatch_image(self.image_id)(
            self.image_id, **self.server_info
        ) as image:
            self.tiler = Tiler.from_hdf5(image, results_path)

        self.cells = Cells.from_source(results_path)

    def random_valid_trap_tp(
        self,
        min_ncells: int = None,
        min_consecutive_tps: int = None,
        label_modulo: int = None,
    ):
        # Call Cells convenience function to pick a random trap and tp
        # containing cells for x cells for y
        return self.cells.random_valid_trap_tp(
            min_ncells=min_ncells,
            min_consecutive_tps=min_consecutive_tps,
        )

    def get_entire_position(self):
        raise (NotImplementedError)

    def get_position_timelapse(self):
        raise (NotImplementedError)

    @property
    def full(self):
        if not hasattr(self, "_full"):
            self._full = {}
        return self._full

    def get_tc(self, tp, channel=None, server_info=None):
        server_info = server_info or self._server_info
        channel = channel or self.tiler.ref_channel

        with self._image_class(self.image_id, **server_info) as image:
            self.tiler.image = image.data
            return self.tiler.get_tc(tp, channel)

    def _find_channels(self, channels: str, guess: bool = True):
        channels = channels or self.tiler.ref_channel
        if isinstance(channels, (int, str)):
            channels = [channels]
        if isinstance(channels[0], str):
            if guess:
                channels = [self.tiler.channels.index(ch) for ch in channels]
            else:
                channels = [
                    re.search(ch, tiler_channels)
                    for ch in channels
                    for tiler_channels in self.tiler.channels
                ]

        return channels

    def get_pos_timepoints(
        self,
        tps: t.Union[int, t.Collection[int]],
        channels: t.Union[str, t.Collection[str]] = None,
        z: int = None,
        server_info=None,
    ):

        if tps and not isinstance(tps, t.Collection):
            tps = range(tps)

        # TODO add support for multiple channels or refactor
        if channels and not isinstance(channels, t.Collection):
            channels = [channels]

        if z is None:
            z = 0

        server_info = server_info or self.server_info
        channels = 0 or self._find_channels(channels)
        z = z or self.tiler.ref_z

        ch_tps = [(channels[0], tp) for tp in tps]
        with self._image_class(self.image_id, **server_info) as image:
            self.tiler.image = image.data
            for ch, tp in ch_tps:
                if (ch, tp) not in self.full:
                    self.full[(ch, tp)] = self.tiler.get_tiles_timepoint(
                        tp, channels=[ch], z=[z]
                    )[:, 0, 0, ..., z]
            requested_trap = {tp: self.full[(ch, tp)] for ch, tp in ch_tps}

            return requested_trap

    def get_labelled_trap(
        self,
        trap_id: int,
        tps: t.Union[range, t.Collection[int]],
        channels=None,
        concatenate=True,
        **kwargs,
    ) -> t.Tuple[np.array]:
        """
        Core method to fetch traps and labels together
        """
        imgs = self.get_pos_timepoints(tps, channels=channels, **kwargs)
        imgs_list = [x[trap_id] for x in imgs.values()]
        outlines = [
            self.cells.at_time(tp, kind="edgemask").get(trap_id, [])
            for tp in tps
        ]
        lbls = [self.cells.labels_at_time(tp).get(trap_id, []) for tp in tps]
        lbld_outlines = [
            np.stack([mask * lbl for mask, lbl in zip(maskset, lblset)]).max(
                axis=0
            )
            if len(lblset)
            else np.zeros_like(imgs_list[0]).astype(bool)
            for maskset, lblset in zip(outlines, lbls)
        ]
        if concatenate:
            lbld_outlines = np.concatenate(lbld_outlines, axis=1)
            imgs_list = np.concatenate(imgs_list, axis=1)
        return lbld_outlines, imgs_list

    def get_images(self, trap_id, trange, channels, **kwargs):
        """
        Wrapper to fetch images
        """
        out = None
        imgs = {}

        for ch in self._find_channels(channels):
            out, imgs[ch] = self.get_labelled_trap(
                trap_id, trange, channels=[ch], **kwargs
            )
        return out, imgs

    def plot_labelled_trap(
        self,
        trap_id: int,
        channels,
        trange: t.Union[range, t.Collection[int]],
        remove_axis: bool = False,
        savefile: str = None,
        skip_outlines: bool = False,
        norm: str = None,
        ncols: int = None,
        local_colours: bool = True,
        img_plot_kwargs: dict = {},
        lbl_plot_kwargs: dict = {"alpha": 0.8},
        **kwargs,
    ):
        """Wrapper to plot time-lapses of individual traps

        Use Cells and Tiler to generate images of cells with their resulting
        outlines.

        Parameters
        ----------
        trap_id : int
            Identifier of trap
        channels : Union[str, int]
            Channels to use
        trange : t.Union[range, t.Collection[int]]
            Range or collection indicating the time-points to use.
        remove_axis : bool
            None, "off", or "x". Determines whether to remove the x-axis, both
            axes or none.
        savefile : str
            Saves file to a location.
        skip_outlines : bool
            Do not add overlay with outlines
        norm : str
            Normalise signals
        ncols : int
            Number of columns to plot.
        local_colours : bool
            Bypass label indicators to guarantee that colours are not repeated
            (TODO implement)
        img_plot_kwargs : dict
            Arguments to pass to plt.imshow used for images.
        lbl_plot_kwargs : dict
            Keyword arguments to pass to label plots.
        **kwargs : dict
            Additional keyword arguments passed to ImageViewer.get_images.

        Examples
        --------
        FIXME: Add docs.

        """
        if ncols is None:
            ncols = len(trange)
        nrows = int(np.ceil(len(trange) / ncols))
        width = self.tiler.tile_size * ncols

        out, images = self.get_images(trap_id, trange, channels, **kwargs)

        # dilation makes outlines easier to see
        out = dilation(out).astype(float)
        out[out == 0] = np.nan

        channel_labels = [
            self.tiler.channels[ch] if isinstance(ch, int) else ch
            for ch in channels
        ]

        assert not norm or norm in (
            "l1",
            "l2",
            "max",
        ), "Invalid norm argument."

        if norm and norm in ("l1", "l2", "max"):
            images = {k: stretch_image(v) for k, v in images.items()}

        images = [concat_pad(img, width, nrows) for img in images.values()]
        # TODO convert to RGB to draw fluorescence with colour
        tiled_imgs = {}
        tiled_imgs["img"] = np.concatenate(images, axis=0)
        tiled_imgs["cell_labels"] = np.concatenate(
            [concat_pad(out, width, nrows) for _ in images], axis=0
        )

        custom_imshow(
            tiled_imgs["img"],
            **img_plot_kwargs,
        )
        custom_imshow(
            tiled_imgs["cell_labels"],
            cmap=sns.color_palette("Paired", as_cmap=True),
            **lbl_plot_kwargs,
        )

        if remove_axis is True:
            plt.axis("off")
        elif remove_axis == "x":
            plt.tick_params(
                axis="x",
                which="both",
                bottom=False,
                top=False,
                labelbottom=False,
            )

        if remove_axis != "True":
            plt.yticks(
                ticks=[
                    (i * self.tiler.tile_size * nrows)
                    + self.tiler.tile_size * nrows / 2
                    for i in range(len(channels))
                ],
                labels=channel_labels,
            )

        if not remove_axis:
            xlabels = (
                ["+ {} ".format(i) for i in range(ncols)]
                if nrows > 1
                else list(trange)
            )
            plt.xlabel("Time-point")

            plt.xticks(
                ticks=[self.tiler.tile_size * (i + 0.5) for i in range(ncols)],
                labels=xlabels,
            )

        if not np.any(out):
            print("ImageViewer:Warning:No cell outlines found")

        if savefile:
            plt.savefig(savefile, bbox_inches="tight", dpi=300)
            plt.close()
        else:
            plt.show()


def concat_pad(a: np.array, width, nrows):
    """
    Melt an array into having multiple blocks as rows
    """
    return np.concatenate(
        np.array_split(
            np.pad(
                a,
                # ((0, 0), (0, width - (a.shape[1] % width))),
                ((0, 0), (0, a.shape[1] % width)),
                constant_values=np.nan,
            ),
            nrows,
            axis=1,
        )
    )



"""
Base functions to extract information from a single cell.

These functions are automatically read by extractor.py, and
so can only have the cell_mask and trap_image as inputs. They
must return only one value.

They assume that there are no NaNs in the image.

We use the module bottleneck when it performs faster than numpy:
- Median
- values containing NaNs (but we make sure this does not happen)
"""
import math
import typing as t

import bottleneck as bn
import numpy as np
from scipy import ndimage


def area(cell_mask) -> int:
    """
    Find the area of a cell mask.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    """
    return np.sum(cell_mask)


def eccentricity(cell_mask) -> float:
    """
    Find the eccentricity using the approximate major and minor axes.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    """
    min_ax, maj_ax = min_maj_approximation(cell_mask)
    return np.sqrt(maj_ax**2 - min_ax**2) / maj_ax


def mean(cell_mask, trap_image) -> float:
    """
    Find the mean of the pixels in the cell.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    trap_image: 2d array
    """
    return np.mean(trap_image[cell_mask])


def median(cell_mask, trap_image) -> int:
    """
    Find the median of the pixels in the cell.

    Parameters
    ----------
    cell_mask: 2d array
         Segmentation mask for the cell.
    trap_image: 2d array
    """
    return bn.median(trap_image[cell_mask])


def max2p5pc(cell_mask, trap_image) -> float:
    """
    Find the mean of the brightest 2.5% of pixels in the cell.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    trap_image: 2d array
    """
    # number of pixels in mask
    npixels = np.sum(cell_mask)
    n_top = int(np.ceil(npixels * 0.025))
    # sort pixels in cell and find highest 2.5%
    pixels = trap_image[cell_mask]
    top_values = bn.partition(pixels, len(pixels) - n_top)[-n_top:]
    # find mean of these highest pixels
    return np.mean(top_values)


def max5px(cell_mask, trap_image) -> float:
    """
    Find the mean of the five brightest pixels in the cell.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    trap_image: 2d array
    """
    # sort pixels in cell
    pixels = trap_image[cell_mask]
    top_values = bn.partition(pixels, len(pixels) - 5)[-5:]
    # find mean of five brightest pixels
    max5px = np.mean(top_values)
    return max5px


def std(cell_mask, trap_image):
    """
    Find the standard deviation of the values of the pixels in the cell.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    trap_image: 2d array
    """
    return np.std(trap_image[cell_mask])


def volume(cell_mask) -> float:
    """
    Estimate the volume of the cell.

    Assumes the cell is an ellipsoid with the mask providing
    a cross-section through its median plane.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell.
    """
    min_ax, maj_ax = min_maj_approximation(cell_mask)
    return (4 * np.pi * min_ax**2 * maj_ax) / 3


def conical_volume(cell_mask):
    """
    Estimate the volume of the cell.

    Parameters
    ----------
    cell_mask: 2D array
        Segmentation mask for the cell
    """
    padded = np.pad(cell_mask, 1, mode="constant", constant_values=0)
    nearest_neighbor = (
        ndimage.morphology.distance_transform_edt(padded == 1) * padded
    )
    return 4 * np.sum(nearest_neighbor)


def spherical_volume(cell_mask):
    """
    Estimate the volume of the cell.

    Assumes the cell is a sphere with the mask providing
    a cross-section through its median plane.

    Parameters
    ----------
    cell_mask: 2d array
        Segmentation mask for the cell
    """
    total_area = area(cell_mask)
    r = math.sqrt(total_area / np.pi)
    return (4 * np.pi * r**3) / 3


def min_maj_approximation(cell_mask) -> t.Tuple[int]:
    """
    Find the lengths of the minor and major axes of an ellipse from a cell mask.

    Parameters
    ----------
    cell_mask: 3d array
        Segmentation masks for cells
    """
    # pad outside with zeros so that the distance transforms have no edge artifacts
    padded = np.pad(cell_mask, 1, mode="constant", constant_values=0)
    # get the distance from the edge, masked
    nn = ndimage.morphology.distance_transform_edt(padded == 1) * padded
    # get the distance from the top of the cone, masked
    dn = ndimage.morphology.distance_transform_edt(nn - nn.max()) * padded
    # get the size of the top of the cone (points that are equally maximal)
    cone_top = ndimage.morphology.distance_transform_edt(dn == 0) * padded
    # minor axis = largest distance from the edge of the ellipse
    min_ax = np.round(np.max(nn))
    # major axis = largest distance from the cone top
    # + distance from the center of cone top to edge of cone top
    maj_ax = np.round(np.max(dn) + np.sum(cone_top) / 2)
    return min_ax, maj_ax



# File with defaults for ease of use
import re
import typing as t
from pathlib import PosixPath
import h5py

# should we move these functions here?
from aliby.tile.tiler import find_channel_name


def exparams_from_meta(
    meta: t.Union[dict, PosixPath, str], extras: t.Collection[str] = ["ph"]
):
    """
    Obtain parameters from metadata of the h5 file.

    Compares a list of candidate channels using case-insensitive
    REGEX to identify valid channels.
    """
    meta = meta if isinstance(meta, dict) else load_metadata(meta)
    base = {
        "tree": {"general": {"None": ["area", "volume", "eccentricity"]}},
        "multichannel_ops": {},
    }
    candidate_channels = {
        "Citrine",
        "GFP",
        "GFPFast",
        "mCherry",
        "pHluorin405",
        "pHluorin488",
        "Flavin",
        "Cy5",
        "mKO2",
    }
    default_reductions = {"max"}
    default_metrics = {
        "mean",
        "median",
        "std",
        "imBackground",
        "max5px",
        # "nuc_est_conv",
    }
    # define ratiometric combinations
    # key is numerator and value is denominator
    # add more to support additional channel names
    ratiometric_combinations = {"phluorin405": ("phluorin488", "gfpfast")}
    default_reduction_metrics = {
        r: default_metrics for r in default_reductions
    }
    # default_rm["None"] = ["nuc_conv_3d"] # Uncomment this to add nuc_conv_3d (slow)
    extant_fluorescence_ch = []
    for av_channel in candidate_channels:
        # find matching channels in metadata
        found_channel = find_channel_name(meta.get("channels", []), av_channel)
        if found_channel is not None:
            extant_fluorescence_ch.append(found_channel)
    for ch in extant_fluorescence_ch:
        base["tree"][ch] = default_reduction_metrics
    base["sub_bg"] = extant_fluorescence_ch
    # additional extraction defaults if the channels are available
    if "ph" in extras:
        # SWAINLAB specific names
        # find first valid combination of ratiometric fluorescence channels
        numerator_channel, denominator_channel = (None, None)
        for ch1, chs2 in ratiometric_combinations.items():
            found_channel1 = find_channel_name(extant_fluorescence_ch, ch1)
            if found_channel1 is not None:
                numerator_channel = found_channel1
                for ch2 in chs2:
                    found_channel2 = find_channel_name(
                        extant_fluorescence_ch, ch2
                    )
                    if found_channel2:
                        denominator_channel = found_channel2
                        break
        # if two compatible ratiometric channels are available
        if numerator_channel is not None and denominator_channel is not None:
            sets = {
                b + a: (x, y)
                for a, x in zip(
                    ["", "_bgsub"],
                    (
                        [numerator_channel, denominator_channel],
                        [
                            f"{numerator_channel}_bgsub",
                            f"{denominator_channel}_bgsub",
                        ],
                    ),
                )
                for b, y in zip(["em_ratio", "gsum"], ["div0", "add"])
            }
            for i, v in sets.items():
                base["multichannel_ops"][i] = [
                    *v,
                    default_reduction_metrics,
                ]
    return base


def load_metadata(file: t.Union[str, PosixPath], group="/"):
    """Get meta data from an h5 file."""
    with h5py.File(file, "r") as f:
        meta = dict(f[group].attrs.items())
    return meta



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class crosscorrParameters(ParametersABC):
    """
    Parameters for the 'crosscorr' process.

    Attributes
    ----------
    normalised: boolean (optional)
        If True, normalise the result for each replicate by the standard
        deviation over time for that replicate.
    only_pos: boolean (optional)
        If True, return results only for positive lags.
    """

    _defaults = {"normalised": True, "only_pos": False}


class crosscorr(PostProcessABC):
    """ """

    def __init__(self, parameters: crosscorrParameters):
        super().__init__(parameters)

    def run(self, trace_dfA: pd.DataFrame, trace_dfB: pd.DataFrame = None):
        """Calculates normalised cross-correlations as a function of lag.

        Calculates normalised auto- or cross-correlations as a function of lag.
        Lag is given in multiples of the unknown time interval between data points.

        Normalisation is by the product of the standard deviation over time for
        each replicate for each variable.

        For the cross-correlation between sA and sB, the closest peak to zero lag
        should in the positive lags if sA is delayed compared to signal B and in the
        negative lags if sA is advanced compared to signal B.

        Parameters
        ----------
        trace_dfA: dataframe
            An array of signal values, with each row a replicate measurement
            and each column a time point.
        trace_dfB: dataframe (required for cross-correlation only)
            An array of signal values, with each row a replicate measurement
            and each column a time point.
        normalised: boolean (optional)
            If True, normalise the result for each replicate by the standard
            deviation over time for that replicate.
        only_pos: boolean (optional)
            If True, return results only for positive lags.

        Returns
        -------
        corr: dataframe
            An array of the correlations with each row the result for the
            corresponding replicate and each column a time point
        lags: array
            A 1D array of the lags in multiples of the unknown time interval

        Example
        -------
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import pandas as pd
        >>> from postprocessor.core.multisignal.crosscorr import crosscorr

        Define a sine signal with 200 time points and 333 replicates

        >>> t = np.linspace(0, 4, 200)
        >>> ts = np.tile(t, 333).reshape((333, 200))
        >>> s = 3*np.sin(2*np.pi*ts + 2*np.pi*np.random.rand(333, 1))
        >>> s_df = pd.DataFrame(s)

        Find and plot the autocorrelaton

        >>> ac = crosscorr.as_function(s_df)
        >>> plt.figure()
        >>> plt.plot(ac.columns, ac.mean(axis=0, skipna=True))
        >>> plt.show()
        """

        df = (
            trace_dfA.copy()
            if type(trace_dfA) == pd.core.frame.DataFrame
            else None
        )
        # convert from aliby dataframe to arrays
        trace_A = trace_dfA.to_numpy()
        # number of replicates
        n_replicates = trace_A.shape[0]
        # number of time points
        n_tps = trace_A.shape[1]
        # deviation from the mean, where the mean is calculated over replicates
        # at each time point, which allows for non-stationary behaviour
        dmean_A = trace_A - np.nanmean(trace_A, axis=0).reshape((1, n_tps))
        # standard deviation over time for each replicate
        stdA = np.sqrt(
            np.nanmean(dmean_A**2, axis=1).reshape((n_replicates, 1))
        )
        if trace_dfB is not None:
            trace_B = trace_dfB.to_numpy()
            # cross correlation
            dmean_B = trace_B - np.nanmean(trace_B, axis=0).reshape((1, n_tps))
            stdB = np.sqrt(
                np.nanmean(dmean_B**2, axis=1).reshape((n_replicates, 1))
            )
        else:
            # auto correlation
            dmean_B = dmean_A
            stdB = stdA
        # calculate correlation
        # lag r runs over positive lags
        pos_corr = np.nan * np.ones(trace_A.shape)
        for r in range(n_tps):
            prods = [
                dmean_A[:, lagtime] * dmean_B[:, lagtime + r]
                for lagtime in range(n_tps - r)
            ]
            pos_corr[:, r] = np.nanmean(prods, axis=0)
        # lag r runs over negative lags
        # use corr_AB(-k) = corr_BA(k)
        neg_corr = np.nan * np.ones(trace_A.shape)
        for r in range(n_tps):
            prods = [
                dmean_B[:, lagtime] * dmean_A[:, lagtime + r]
                for lagtime in range(n_tps - r)
            ]
            neg_corr[:, r] = np.nanmean(prods, axis=0)
        if self.normalised:
            # normalise by standard deviation
            pos_corr = pos_corr / stdA / stdB
            neg_corr = neg_corr / stdA / stdB
        # combine lags
        lags = np.arange(-n_tps + 1, n_tps)
        corr = np.hstack((np.flip(neg_corr[:, 1:], axis=1), pos_corr))
        if self.only_pos:
            return pd.DataFrame(
                corr[:, int(lags.size / 2) :],
                index=df.index,
                columns=lags[int(lags.size / 2) :],
            )
        else:
            return pd.DataFrame(corr, index=df.index, columns=lags)



from itertools import cycle

import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class aggregateParameters(ParametersABC):
    """
    Parameters
    reduction: str to be passed to a dataframe for collapsing across columns
    """

    _defaults = {
        "reductions": ["mean", "median", "max"],
        "axis": 1,
        "ranges": None,
    }


class aggregate(PostProcessABC):
    """
    Aggregate multiple datasets for cell-to-cell feature comparison.
    """

    def __init__(self, parameters: aggregateParameters):
        super().__init__(parameters)

    def run(self, signals):
        index = signals[0].index
        for s in signals[0:]:
            index = index.intersection(s.index)

        tmp_signals = [s.loc[index] for s in signals]
        for i, s in enumerate(signals):
            tmp_signals[i].name = s.name
        signals = tmp_signals

        assert len(signals), "Signals is empty"

        bad_words = {
            "postprocessing",
            "extraction",
            "None",
            "np_max",
            "",
        }

        def get_keywords(df):
            return [
                ind
                for item in df.name.split("/")
                for ind in item.split("/")
                if ind not in bad_words
            ]

        colnames = [
            "_".join(get_keywords(s) + [red])
            for s in signals
            for red in self.parameters.reductions
        ]
        concat = pd.concat(
            [
                getattr(signal, red)(axis=self.parameters.axis)
                for signal in signals
                for red in self.parameters.reductions
            ],
            names=signals[0].index.names,
            axis=self.parameters.axis,
        )
        if self.parameters.axis:
            concat.columns = colnames
        else:
            concat.columns = pd.MultiIndex.from_product(
                (
                    colnames,
                    [
                        "_".join((str(start), str(stop)))
                        for x in self.parameters.ranges
                        for start, stop in x
                    ],
                )
            )

        return concat



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm, ticker

from postprocessor.core.processes.standardscaler import standardscaler
from postprocessor.routines.plottingabc import BasePlotter


class _HeatmapPlotter(BasePlotter):
    """Draw heatmap"""

    def __init__(
        self,
        trace_df,
        trace_name,
        buddings_df,
        cmap,
        unit_scaling,
        xtick_step,
        scale,
        robust,
        xlabel,
        ylabel,
        cbarlabel,
        plot_title,
    ):
        super().__init__(trace_name, unit_scaling, xlabel, plot_title)
        # Define attributes from arguments
        self.trace_df = trace_df
        self.buddings_df = buddings_df
        self.cmap = cmap
        self.xtick_step = xtick_step
        self.scale = scale
        self.robust = robust

        # Define some labels
        self.cbarlabel = cbarlabel
        self.ylabel = ylabel

        # Scale
        if self.scale:
            self.trace_scaled = standardscaler.as_function(self.trace_df)
        else:
            self.trace_scaled = self.trace_df

        # If robust, redefine colormap scale to remove outliers
        if self.robust:
            self.vmin = np.nanpercentile(self.trace_scaled, 2)
            self.vmax = np.nanpercentile(self.trace_scaled, 98)
            # Make axes even
            if self.scale:
                if np.abs(self.vmin) > np.abs(self.vmax):
                    self.vmax = -self.vmin
                else:
                    self.vmin = -self.vmax
        else:
            self.vmin = None
            self.vmax = None

        # Define horizontal axis ticks and labels
        # hacky! -- redefine column names
        trace_df.columns = trace_df.columns * self.unit_scaling
        self.fmt = ticker.FuncFormatter(
            lambda x, pos: "{0:g}".format(x * self.unit_scaling)
        )

    def plot(self, ax, cax):
        """Draw the heatmap on the provided Axes."""
        super().plot(ax)
        ax.xaxis.set_major_formatter(self.fmt)
        # Draw trace heatmap
        trace_heatmap = ax.imshow(
            self.trace_scaled,
            cmap=self.cmap,
            interpolation="none",
            vmin=self.vmin,
            vmax=self.vmax,
        )
        # Horizontal axis labels as multiples of xtick_step, taking
        # into account unit scaling
        ax.xaxis.set_major_locator(
            ticker.MultipleLocator(self.xtick_step / self.unit_scaling)
        )
        # Overlay buddings, if present
        if self.buddings_df is not None:
            # Must be masked array for transparency
            buddings_array = self.buddings_df.to_numpy()
            buddings_heatmap_mask = np.ma.masked_where(
                buddings_array == 0, buddings_array
            )
            # Overlay
            ax.imshow(
                buddings_heatmap_mask,
                interpolation="none",
            )
        # Draw colour bar
        ax.figure.colorbar(
            mappable=trace_heatmap, cax=cax, ax=ax, label=self.cbarlabel
        )


def heatmap(
    trace_df,
    trace_name,
    buddings_df=None,
    cmap=cm.RdBu,
    unit_scaling=1,
    xtick_step=60,
    scale=True,
    robust=True,
    xlabel="Time (min)",
    ylabel="Cell",
    cbarlabel="Normalised fluorescence (AU)",
    plot_title="",
    ax=None,
    cbar_ax=None,
):
    """Draw heatmap from an array of time series of traces

    Parameters
    ----------
    trace_df : pandas.DataFrame
        Time series of traces (rows = cells, columns = time points).
    trace_name : string
        Name of trace being plotted, e.g. 'flavin'.
    buddings_df : pandas.DataFrame
        Birth mask (rows = cells, columns = time points).  Elements should be
        0 or 1.
    cmap : matplotlib ColorMap
        Colour map for heatmap.
    unit_scaling : int or float
        Unit scaling factor, e.g. 1/60 to convert minutes to hours.
    xtick_step : int or float
        Interval length, in unit time, to draw x axis ticks.
    scale : bool
        Whether to use standard scaler to scale the trace time series.
    robust : bool
        If True, the colour map range is computed with robust quantiles instead
        of the extreme values.
    xlabel : string
        x axis label.
    ylabel : string
        y axis label.
    cbarlabel : string
        Colour bar label.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.
    cbar_ax : matplotlib Axes
        Axes in which to draw the colour bar, otherwise take space from the main
        Axes.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the heatmap.

    Examples
    --------
    FIXME: Add docs.

    """
    plotter = _HeatmapPlotter(
        trace_df,
        trace_name,
        buddings_df,
        cmap,
        unit_scaling,
        xtick_step,
        scale,
        robust,
        xlabel,
        ylabel,
        cbarlabel,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax, cbar_ax)
    return ax



from itertools import takewhile
from typing import Dict, List, Union

import h5py
import numpy as np
import pandas as pd
from tqdm import tqdm

from agora.abc import ParametersABC, ProcessABC
from agora.io.cells import Cells
from agora.io.signal import Signal
from agora.io.writer import Writer
from postprocessor.core.abc import get_parameters, get_process
from postprocessor.core.lineageprocess import LineageProcessParameters
from postprocessor.core.reshapers.merger import Merger, MergerParameters
from postprocessor.core.reshapers.picker import Picker, PickerParameters


class PostProcessorParameters(ParametersABC):
    """
    Anthology of parameters used for postprocessing
    :merger:
    :picker: parameters for picker
    :processes: list processes:[objectives], 'processes' are defined in ./processes/
        while objectives are relative or absolute paths to datasets. If relative paths the
        post-processed addresses are used. The order of processes matters.

    """

    def __init__(
        self,
        targets={},
        param_sets={},
        outpaths={},
    ):
        self.targets: Dict = targets
        self.param_sets: Dict = param_sets
        self.outpaths: Dict = outpaths

    def __getitem__(self, item):
        return getattr(self, item)

    @classmethod
    def default(cls, kind=[]):
        """Sequential postprocesses to be operated"""
        targets = {
            "prepost": {
                "merger": "/extraction/general/None/area",
                "picker": ["/extraction/general/None/area"],
            },
            "processes": [
                [
                    "buddings",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "dsignal",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "bud_metric",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "dsignal",
                    [
                        "/postprocessing/bud_metric/extraction_general_None_volume",
                    ],
                ],
            ],
        }
        param_sets = {
            "prepost": {
                "merger": MergerParameters.default(),
                "picker": PickerParameters.default(),
            }
        }
        outpaths = {}
        outpaths["aggregate"] = "/postprocessing/experiment_wide/aggregated/"

        if "ph_batman" in kind:
            targets["processes"]["dsignal"].append(
                [
                    "/extraction/em_ratio/np_max/mean",
                    "/extraction/em_ratio/np_max/median",
                    "/extraction/em_ratio_bgsub/np_max/mean",
                    "/extraction/em_ratio_bgsub/np_max/median",
                ]
            )
            targets["processes"]["aggregate"].append(
                [
                    [
                        "/extraction/em_ratio/np_max/mean",
                        "/extraction/em_ratio/np_max/median",
                        "/extraction/em_ratio_bgsub/np_max/mean",
                        "/extraction/em_ratio_bgsub/np_max/median",
                        "/extraction/gsum/np_max/median",
                        "/extraction/gsum/np_max/mean",
                    ]
                ],
            )

        return cls(targets=targets, param_sets=param_sets, outpaths=outpaths)


class PostProcessor(ProcessABC):
    def __init__(self, filename, parameters):
        super().__init__(parameters)
        self._filename = filename
        self._signal = Signal(filename)
        self._writer = Writer(filename)

        dicted_params = {
            i: parameters["param_sets"]["prepost"][i]
            for i in ["merger", "picker"]
        }

        for k in dicted_params.keys():
            if not isinstance(dicted_params[k], dict):
                dicted_params[k] = dicted_params[k].to_dict()

        self.merger = Merger(
            MergerParameters.from_dict(dicted_params["merger"])
        )

        self.picker = Picker(
            PickerParameters.from_dict(dicted_params["picker"]),
            cells=Cells.from_source(filename),
        )
        self.classfun = {
            process: get_process(process)
            for process, _ in parameters["targets"]["processes"]
        }
        self.parameters_classfun = {
            process: get_parameters(process)
            for process, _ in parameters["targets"]["processes"]
        }
        self.targets = parameters["targets"]

    def run_prepost(self):
        # TODO Split function
        """Important processes run before normal post-processing ones"""

        merge_events = self.merger.run(
            self._signal[self.targets["prepost"]["merger"]]
        )

        prev_idchanges = self._signal.get_merges()

        changes_history = list(prev_idchanges) + [
            np.array(x) for x in merge_events
        ]
        self._writer.write("modifiers/merges", data=changes_history)

        # TODO Remove this once test is wriiten for consecutive postprocesses
        with h5py.File(self._filename, "a") as f:
            if "modifiers/picks" in f:
                del f["modifiers/picks"]

        indices = self.picker.run(
            self._signal[self.targets["prepost"]["picker"][0]]
        )

        combined_idx = ([], [], [])
        trap, mother, daughter = combined_idx

        lineage = self.picker.cells.mothers_daughters

        if lineage.any():
            trap, mother, daughter = lineage.T
            combined_idx = np.vstack((trap, mother, daughter))

        trap_mother = np.vstack((trap, mother)).T
        trap_daughter = np.vstack((trap, daughter)).T

        multii = pd.MultiIndex.from_arrays(
            combined_idx,
            names=["trap", "mother_label", "daughter_label"],
        )
        self._writer.write(
            "postprocessing/lineage",
            data=multii,
            overwrite="overwrite",
        )

        # apply merge to mother-trap_daughter
        moset = set([tuple(x) for x in trap_mother])
        daset = set([tuple(x) for x in trap_daughter])
        picked_set = set([tuple(x) for x in indices])

        with h5py.File(self._filename, "a") as f:
            merge_events = f["modifiers/merges"][()]
        multii = pd.MultiIndex(
            [[], [], []],
            [[], [], []],
            names=["trap", "mother_label", "daughter_label"],
        )
        self.lineage_merged = multii

        if merge_events.any():

            def search(a, b):
                return np.where(
                    np.in1d(
                        np.ravel_multi_index(a.T, a.max(0) + 1),
                        np.ravel_multi_index(b.T, a.max(0) + 1),
                    )
                )

            for target, source in merge_events:
                if (
                    tuple(source) in moset
                ):  # update mother to lowest positive index among the two
                    mother_ids = search(trap_mother, source)
                    trap_mother[mother_ids] = (
                        target[0],
                        self.pick_mother(
                            trap_mother[mother_ids][0][1], target[1]
                        ),
                    )
                if tuple(source) in daset:
                    trap_daughter[search(trap_daughter, source)] = target
                if tuple(source) in picked_set:
                    indices[search(indices, source)] = target

            self.lineage_merged = pd.MultiIndex.from_arrays(
                np.unique(
                    np.append(
                        trap_mother,
                        trap_daughter[:, 1].reshape(-1, 1),
                        axis=1,
                    ),
                    axis=0,
                ).T,
                names=["trap", "mother_label", "daughter_label"],
            )
        self._writer.write(
            "postprocessing/lineage_merged",
            data=self.lineage_merged,
            overwrite="overwrite",
        )

        self._writer.write(
            "modifiers/picks",
            data=pd.MultiIndex.from_arrays(
                # TODO Check if multiindices are still repeated
                np.unique(indices, axis=0).T if indices.any() else [[], []],
                names=["trap", "cell_label"],
            ),
            overwrite="overwrite",
        )

    @staticmethod
    def pick_mother(a, b):
        """Update the mother id following this priorities:

        The mother has a lower id
        """
        x = max(a, b)
        if min([a, b]):
            x = [a, b][np.argmin([a, b])]
        return x

    def run(self):
        # TODO Documentation :) + Split
        self.run_prepost()

        for process, datasets in tqdm(self.targets["processes"]):
            if process in self.parameters["param_sets"].get(
                "processes", {}
            ):  # If we assigned parameters
                parameters = self.parameters_classfun[process](
                    self.parameters[process]
                )

            else:
                parameters = self.parameters_classfun[process].default()

            if isinstance(parameters, LineageProcessParameters):
                lineage = self._signal.lineage(
                    # self.parameters.lineage_location
                )
                loaded_process = self.classfun[process](parameters)
                loaded_process.lineage = lineage

            else:
                loaded_process = self.classfun[process](parameters)

            for dataset in datasets:
                if isinstance(dataset, list):  # multisignal process
                    signal = [self._signal[d] for d in dataset]
                elif isinstance(dataset, str):
                    signal = self._signal[dataset]
                else:
                    raise ("Incorrect dataset")

                if len(signal):
                    result = loaded_process.run(signal)
                else:
                    result = pd.DataFrame(
                        [], columns=signal.columns, index=signal.index
                    )
                    result.columns.names = ["timepoint"]

                if process in self.parameters["outpaths"]:
                    outpath = self.parameters["outpaths"][process]
                elif isinstance(dataset, list):
                    # If no outpath defined, place the result in the minimum common
                    # branch of all signals used
                    prefix = "".join(
                        c[0]
                        for c in takewhile(
                            lambda x: all(x[0] == y for y in x), zip(*dataset)
                        )
                    )
                    outpath = (
                        prefix
                        + "_".join(  # TODO check that it always finishes in '/'
                            [
                                d[len(prefix) :].replace("/", "_")
                                for d in dataset
                            ]
                        )
                    )
                elif isinstance(dataset, str):
                    outpath = dataset[1:].replace("/", "_")
                else:
                    raise ("Outpath not defined", type(dataset))

                if process not in self.parameters["outpaths"]:
                    outpath = "/postprocessing/" + process + "/" + outpath

                if isinstance(result, dict):  # Multiple Signals as output
                    for k, v in result.items():
                        self.write_result(
                            outpath + f"/{k}",
                            v,
                            metadata={},
                        )
                else:
                    self.write_result(
                        outpath,
                        result,
                        metadata={},
                    )

    def write_result(
        self,
        path: str,
        result: Union[List, pd.DataFrame, np.ndarray],
        metadata: Dict,
    ):
        self._writer.write(path, result, meta=metadata, overwrite="overwrite")
